/* $Id: nntpcl.c 1.4 1996/05/18 21:12:57 cthuang Exp $
 *
 * NNTP client routines
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#ifdef MPTN
#include <utils.h>
#endif
#ifdef __NT__
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif
#include "socket.h"
#include "nntp.h"
#include "nntpcl.h"
#include "souper.h"

#define NNTP_PORT 119

/* size of buffer to store user name */
#define USER_MAX 80

/* Open connection to NNTP server.
 * Return socket handle or -1 on error.
 */
int
nntpConnect (void)
{
    struct servent *sp;
    char buf[BUFSIZ];
    int response;
    int socket;
    short port;

    if (nntpServer == NULL) {
	fprintf(stderr,
	    "Set the NNTPSERVER environment variable to the news host.\n");
	exit(EXIT_FAILURE);
    }

    sp = getservbyname("nntp", "tcp");
    port = (sp == NULL) ? NNTP_PORT : ntohs(sp->s_port);

    socket = Socket(nntpServer, port);
    if (socket < 0) {
	fprintf(stderr, "Cannot connect to news server %s\n", nntpServer);
	return -1;
    }

    if (SockGets(socket, buf, sizeof(buf)) == 0) {
	response = atoi(buf);
	switch (response) {
	case OK_NOPOST:
	    printf("You cannot post articles to the news server %s\n",
		nntpServer);
	    break;

	case OK_CANPOST:
	    break;

	case ERR_ACCESS:
	    printf("You do not have permission to use the news server %s\n",
		nntpServer);
	    return -1;

	default:
	    printf("Unexpected response from news server %s\n", nntpServer);
	    puts(buf);
	    return -1;
	}
    }

    /* This is for INN */
    SockPuts(socket, "mode reader");
    SockGets(socket, buf, sizeof(buf));

    return socket;
}

/* Close NNTP connection. */

void
nntpClose (int socket)
{
    SockPuts(socket, "QUIT");
    SockClose(socket);
}

/* Get authentication information required for the NNTP server.
 * Return TRUE if successful.
 */
static int
getUserPass (char *user, char *pass)
{
    char newsauthFile[FILENAME_MAX], server[USER_MAX], buf[BUFSIZ];
    FILE *inf;

#if defined(__OS2__) || defined(__NT__)
    sprintf(newsauthFile, "%s/newsauth", homeDir);
#else
    sprintf(newsauthFile, "%s/.newsauth", homeDir);
#endif
    if ((inf = fopen(newsauthFile, "r")) == NULL) {
	perror(newsauthFile);
	return 0;
    }

    while (fgets(buf, sizeof(buf), inf)) {
	if (sscanf(buf, "%s %s %s", server, user, pass) != 3)
	    continue;
	if (stricmp(server, nntpServer) == 0) {
	    fclose(inf);
	    return 1;
	}
    }

    fclose(inf);
    fprintf(stderr, "No authorization information for %s in %s\n",
	server, newsauthFile);
    return 0;
}

/* Send an NNTP request.
 * Return TRUE if successful.
 */
int
nntpRequest (int socket, const char *request, char *reply, size_t replyMax)
{
    char buf[BUFSIZ];
    char user[USER_MAX], pass[USER_MAX];
    int code;

    /* Send request. */
    SockPuts(socket, request);
    if (SockGets(socket, reply, replyMax) != 0)
	return 0;
    if (reply[0] == CHAR_FATAL) {
	fprintf(stderr, "%s\n", reply);
	exit(EXIT_FAILURE);
    }

    code = atoi(reply);
    if (code != 480) {
	if (reply[0] == CHAR_ERR && code != 423)
	    fprintf(stderr, "%s\n", reply);
	return reply[0] != CHAR_ERR;
    }

    /* Authentication required for command. */
    if (!getUserPass(user, pass))
	exit(EXIT_FAILURE);

    SockPrintf(socket, "AUTHINFO USER %s\r\n", user);
    if (SockGets(socket, buf, sizeof(buf)) != 0)
	return 0;
    if (atoi(buf) != 381) {
	fprintf(stderr, "%s\n", buf);
	return 0;
    }

    /* Password required. */
    SockPrintf(socket, "AUTHINFO PASS %s\r\n", pass);
    if (SockGets(socket, buf, sizeof(buf)) != 0)
	return 0;
    if (atoi(buf) != 281) {
	fprintf(stderr, "%s\n", buf);
	return 0;
    }
    printf("Sent AUTHINFO USER %s to %s\n", user, nntpServer);

    /* Send request again. */
    SockPuts(socket, request);
    if (SockGets(socket, reply, replyMax) != 0)
	return 0;
    if (reply[0] == CHAR_FATAL) {
	fprintf(stderr, "%s\n", reply);
	exit(EXIT_FAILURE);
    }
    return reply[0] != CHAR_ERR;
}

/* Select newsgroup to read from
 * Return TRUE if successful.
 */
int
nntpGroup (int socket, const char *ngname, ArticleNumber *pLo,
    ArticleNumber *pHi)
{
    char request[REQUEST_MAX], buf[BUFSIZ];
    int count;

    sprintf(request, "GROUP %s", ngname);
    if (nntpRequest(socket, request, buf, sizeof(buf))) {
	sscanf(buf+4, "%d %d %d", &count, pLo, pHi);
	return 1;
    }
    return 0;
}

/* Request overview for the current newsgroup.
 * Return TRUE if successful.
 */
int
nntpXover (int socket, ArticleNumber lo, ArticleNumber hi)
{
    char request[REQUEST_MAX], buf[BUFSIZ];

    if (lo < hi)
	sprintf(request, "XOVER %ld-%ld", lo, hi);
    else
	sprintf(request, "XOVER %ld", lo);

    return nntpRequest(socket, request, buf, sizeof(buf));
}

/* Get next article in group.
 * Return TRUE if successful.
 */
int
nntpNext (int socket, ArticleNumber *pArtNum)
{
    char buf[BUFSIZ];

    if (nntpRequest(socket, "NEXT", buf, sizeof(buf))) {
	sscanf(buf+4, "%ld", pArtNum);
	return 1;
    }
    return 0;
}

/* Get article from server.
 * Return TRUE if successful.
 */
int
nntpArticle (int socket, const char *cmd, ArticleNumber artnum, FILE *outf)
{
    char request[REQUEST_MAX], buf[BUFSIZ];
    char *bufp;

    sprintf(request, "%s %ld\r\n", cmd, artnum);
    if (!nntpRequest(socket, request, buf, sizeof(buf)))
	return 0;

    while (SockGets(socket, buf, sizeof(buf)) == 0) {
	bufp = buf;
	if (buf[0] == '.') {
	    ++bufp;
	    if (buf[1] == '\0')
		break;
	}

	fputs(bufp, outf);
	fputc('\n', outf);
    }

    return 1;
}

/* Get date from server.
 * Return TRUE if successful.
 */
int
nntpDate (int socket, char *dest)
{
    char buf[BUFSIZ];

    if (nntpRequest(socket, "DATE", buf, sizeof(buf))) {
	sscanf(buf+4, "%s", dest);
	return 1;
    }
    return 0;
}

/* Post article to NNTP server.
 * Return TRUE if successful.
*/
int
nntpPost (int socket, FILE *inf, size_t bytes)
{
    char buf[BUFSIZ], *s;
    size_t len, count;
    long offset;

    if (!nntpRequest(socket, "POST", buf, sizeof(buf)))
	return 0;

    offset = ftell(inf);
    count = bytes;
    while (fgets(buf, sizeof(buf), inf) && count > 0) {
	count -= strlen(buf);
	if ((s = strchr(buf, '\n')) != NULL)
	    *s = '\0';
	if (buf[0] == '.')
	    SockWrite(socket, buf, 1);
	SockPrintf(socket, "%s\r\n", buf);
    }
    fseek(inf, offset+bytes, SEEK_SET);

    SockPrintf(socket, ".\r\n");

    if (SockGets(socket, buf, sizeof(buf)) < 0) {
	return 0;
    }
    if (buf[0] != CHAR_OK) {
	if (atoi(buf) == ERR_POSTFAIL) {
	    fprintf(stderr, "Article not accepted by server; not posted.\n");
	}
	fprintf(stderr, "%s\n", buf);
	return 0;
    }

    return 1;
}
