/* $Id: score.c $
 */
#include <limits.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "score.h"
#include "souper.h"

static GroupScore *globalScore;		/* global score rules */
static GroupScore *groupScoreList;	/* list of group specific rules */
static GroupScore *curGroupScore; 	/* current group rules */

/* Create regular expression and compile pattern.
 */
static regexp_t
newRegExp (const char *pattern)
{
    static char translateToLower[256];
    regexp_t re;
    char *msg;
    int i;

    re = (regexp_t)xmalloc(sizeof(struct re_pattern_buffer));
    memset(re, 0, sizeof(struct re_pattern_buffer));

    for (i = 0; i < 256; ++i)
        translateToLower[i] = tolower(i);
    re->translate = translateToLower;

    re_set_syntax(RE_SYNTAX_AWK);
    msg = re_compile_pattern(pattern, strlen(pattern), re);
    if (msg) {
	fprintf(stderr, "Pattern %s\n%s\n", pattern, msg);
	return NULL;
    } else {
	re->fastmap = xmalloc(256);
    }

    return re;
}

/* Create score rule.
 */
static ScoreRule *
newScoreRule (Score score_, int useRe_, char *where_, char *pattern_)
{
    ScoreRule *rule;
    char *s;

    rule = (ScoreRule *)xmalloc(sizeof(ScoreRule));
    rule->next = NULL;
    rule->score = score_;

    if ((s = strchr(where_, ':')) != NULL)
	*s = '\0';
    if (stricmp(where_, "header") == 0)
	rule->header = NULL;
    else
	rule->header = xstrdup(where_);

    if (useRe_) {
	rule->re = newRegExp(pattern_);
	rule->type = (rule->re == NULL) ? INVALID : REGEXP;
	rule->pattern = NULL;
    } else {
	rule->type = STRING;
	rule->pattern = xstrdup(pattern_);
    }

    return rule;
}

/* Read group score rules from file.
 */
static GroupScore *
readGroupScore (FILE *inf, const char *name)
{
    static const char whiteSpace[] = " \t\n";
    GroupScore *groupScore;
    ScoreRule *pLastRule;
    char line[BUFSIZ];

    groupScore = (GroupScore *)xmalloc(sizeof(GroupScore));
    groupScore->next = NULL;
    groupScore->name = xstrdup(name);
    groupScore->rules = NULL;
    groupScore->killThreshold = 0;
    pLastRule = NULL;

    while (fgets(line, sizeof(line), inf)) {
	char *token = strtok(line, whiteSpace);
	if (token == NULL || token[0] == '#')
	    continue;
	if (token[0] == '}')
	    break;

	if (stricmp(token, "killthreshold") == 0) {
	    /* killthreshold score */
	    if ((token = strtok(NULL, whiteSpace)) != 0) {
		groupScore->killThreshold = atoi(token);
	    }

	} else if (token[0] == '-' || token[0] == '+'
	 || (isascii(token[0]) && isdigit(token[0]))) {
	    /* Line starts with a number.  This may be a score rule. */
	    ScoreRule *rule;
	    int useRe;
	    char *where, *pattern;

	    int score = atoi(token);

	    if ((where = strtok(NULL, whiteSpace)) == 0)
		continue;

	    if (stricmp(where, "pattern") == 0) {
		if ((where = strtok(NULL, whiteSpace)) == 0)
		    continue;
		useRe = 1;
	    } else {
		useRe = 0;
	    }

	    if ((pattern = strtok(NULL, "\n")) == 0)
		continue;
	    while (*pattern == ' ' || *pattern == '\t')
		++pattern;

	    if ((rule = newScoreRule(score, useRe, where, pattern)) != NULL) {
		if (pLastRule == NULL)
		    groupScore->rules = rule;
		else
		    pLastRule->next = rule;
		pLastRule = rule;
	    }
	}
    }

    return groupScore;
}

/* Read score file and compile regular expressions.
 * Return TRUE if successful.
 */
int
readScoreFile (void)
{
    char buf[BUFSIZ], name[FILENAME_MAX];
    FILE *inf;
    GroupScore *pGroup, *pLastGroup;
    char ok;

    globalScore = groupScoreList = NULL;

    if ((inf = fopen(scoreFile, "r")) == NULL) {
	return 0;
    }

    pLastGroup = NULL;
    ok = 1;

    /* Read newsgroup name. */
    while (ok && fscanf(inf, "%s", name) == 1) {
	/* Read opening brace. */
	fscanf(inf, "%s", buf);
	if (buf[0] != '{' || buf[1] != '\0') {
	    ok = 0;
	    break;
	}

	pGroup = readGroupScore(inf, name);
	if (stricmp(name, "all") == 0) {
	    globalScore = pGroup;
	} else {
	    if (pLastGroup == NULL)
		groupScoreList = pGroup;
	    else
		pLastGroup->next = pGroup;
	    pLastGroup = pGroup;
	}
    }
    fclose(inf);

    if (!ok)
	fprintf(stderr, "%s: error in score file %s\n", progname, scoreFile);
    return ok;
}

/* Set current group score.
 * Return TRUE if score processing is enabled.
 */
int
scoreGroup (const char *name)
{
    GroupScore *p;

    for (p = groupScoreList; p != NULL; p = p->next) {
	if (stricmp(p->name, name) == 0) {
	    curGroupScore = p;
	    return 1;
	}
    }
    curGroupScore = NULL;
    return globalScore != NULL || maxLines > 0;
}

/* Return current group kill threshold.
 */
Score
getKillThreshold (void)
{
    if (curGroupScore)
	return curGroupScore->killThreshold;
    if (globalScore)
	return globalScore->killThreshold;
    return 0;
}

typedef unsigned char uchar;

#define LARGE 32767             /* flag for last character match    */

static int patlen;              /* # chars in pattern               */
static int skip[UCHAR_MAX+1];   /* skip-ahead count for test chars  */
static int skip2;               /* skip-ahead after non-match with
                                ** matching final character         */
static uchar *pat = 0;          /* uppercase copy of pattern        */

/* Calculate the skip array for the given pattern.
 */
static void
bmhi_init (const char *pattern)
{
      int i, lastpatchar;

      patlen = strlen(pattern);

      /* Make uppercase copy of pattern */

      free(pat);
      pat = xmalloc(patlen);
      for (i=0; i < patlen; i++)
            pat[i] = toupper(pattern[i]);

      /* initialize skip array */

      for ( i = 0; i < UCHAR_MAX + 1; ++i )
            skip[i] = patlen;
      for ( i = 0; i < patlen - 1; ++i )
      {
            skip[        pat[i] ] = patlen - i - 1;
            skip[tolower(pat[i])] = patlen - i - 1;
      }
      lastpatchar = pat[patlen - 1];
      skip[        lastpatchar ] = LARGE;
      skip[tolower(lastpatchar)] = LARGE;
      skip2 = patlen;                     /* Horspool's fixed second shift */
      for (i = 0; i < patlen - 1; ++i)
      {
            if ( pat[i] == lastpatchar )
                  skip2 = patlen - i - 1;
      }
}

static const char *
bmhi_search (const char *string, const char *pattern)
{
      int i, j, stringlen = strlen(string);
      const char *s;

      bmhi_init(pattern);
      i = patlen - 1 - stringlen;
      if (i >= 0)
      {
            return 0;
      }
      string += stringlen;
      for ( ;; )
      {
            while ( i <= 0 && (i += skip[((uchar*)string)[i]]) < 0 )
                ;
            if (i < (LARGE - stringlen))
                  return 0;
            i -= LARGE;
            j = patlen - 1;
            s = string + (i - j);
            while ( --j >= 0 && toupper(s[j]) == pat[j] )
                  ;
            if ( j >= 0 )
                  i += skip2;
            else
                return s;
      }
}

/* Return TRUE if the line matches the pattern.
 */
static int
matchScoreRule (ScoreRule *rule, const char *line)
{
    size_t headerLen, len;

    if (rule->header) {
	/* Look for the pattern is a specific header. */
	headerLen = strlen(rule->header);
	if (!isHeader(line, rule->header, headerLen))
	    return 0;
	line += headerLen + 2;
    }

    switch (rule->type) {
    case STRING:
	return bmhi_search(line, rule->pattern) != 0;
    case REGEXP:
	len = strlen(line);
	return re_search(rule->re, line, len, 0, len, NULL) >= 0;
    default:
	return 0;
    }
}

/* Process score rules on the line.
 */
static Score
calcGroupScore (GroupScore *groupScore, const char *line)
{
    ScoreRule *rule;
    Score score = 0;

    for (rule = groupScore->rules; rule != NULL; rule = rule->next) {
	if (matchScoreRule(rule, line)) {
	    score += rule->score;
	    printf("%s: score %d %s\n", progname, rule->score, line);
	}
    }
    return score;
}

/* Check if line matches score criteria.
 * Return score
 */
Score
scoreLine (const char *line)
{
    Score score = 0;

    if (maxLines > 0 && strnicmp(line, "Lines: ", 7) == 0) {
	int numLines = atoi(line + 7);
	if (numLines > maxLines) {
	    score -= numLines;
	    printf("%s: score %d %s\n", progname, -numLines, line);
	}
    }

    if (globalScore)
	score += calcGroupScore(globalScore, line);
    if (curGroupScore)
	score += calcGroupScore(curGroupScore, line);
    return score;
}
